/* Remove sphinx-gallery note at the top of generated rst files*/
div.sphx-glr-download-link-note {
    display: none;
}

/* Main index page overview cards */
.sd-card .sd-card-img-top {
  height: 60px;
  width: 60px;
  margin-left: auto;
  margin-right: auto;
  margin-top: 10px;
}

.navbar-brand {
    padding: .65rem 0;
}

/* Show inverted colors in svg icons, if using a dark theme */
html[data-theme=dark] .sd-card img[src*='.svg'] {
  filter: invert(0.82) brightness(0.8) contrast(1.2);
}

html[data-theme=dark] .bd-content .nboutput .output_area .widget-subarea, html[data-theme=dark] .bd-content .nboutput .output_area.rendered_html:not(:has(table.dataframe)) {
    background-color: var(--pst-color-background);
    border-radius: .25rem;
    color: var(--pst-color-on-background);
    padding: .5rem;
}
